/*
 * Copyright 2010-2024 OpenXcom Developers.
 *
 * This file is part of OpenXcom.
 *
 * OpenXcom is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * OpenXcom is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with OpenXcom.  If not, see <http://www.gnu.org/licenses/>.
 */

 // throw exceptions more often
#define RYML_DEFAULT_CALLBACK_USES_EXCEPTIONS
// when using a single header file for the whole rapidayaml, this needs to be defined once, in one of the cpp files
#define RYML_SINGLE_HDR_DEFINE_NOW
// include extra functionality for interaction with std types
#define _RYML_STD_HPP_

#include "Yaml.h"
#include <string>

namespace OpenXcom
{

namespace YAML
{


YamlString::YamlString(std::string yamlString) : yaml(yamlString) {}

YamlNodeReader::YamlNodeReader()
	: _node(ryml::ConstNodeRef(nullptr, ryml::NONE)), _invalid(true), _index(nullptr)
{
}

YamlNodeReader::YamlNodeReader(const ryml::ConstNodeRef& node)
	: _node(node), _invalid(node.invalid()), _index(nullptr)
{
}

YamlNodeReader::YamlNodeReader(const ryml::ConstNodeRef& node, bool useIndex)
	: _node(node), _invalid(node.invalid()), _index(nullptr)
{
	if (!useIndex)
		return;
	// build and use an index to avoid [] operator's O(n) complexity
	_index = new std::unordered_map<ryml::csubstr, ryml::id_type>(_node.num_children());
	for (const ryml::ConstNodeRef& childNode : _node.children())
		(*_index)[childNode.key()] = childNode.id();
}

YamlNodeReader::~YamlNodeReader()
{
	if (_index)
		delete _index;
}

const YamlNodeReader YamlNodeReader::useIndex() const
{
	return YamlNodeReader(_node, true);
}

size_t YamlNodeReader::childrenCount() const
{
	if (_invalid)
		return 0;
	return _node.num_children();
}

ryml::ConstNodeRef YamlNodeReader::getChildNode(ryml::csubstr key) const
{
	if (_invalid)
		return ryml::ConstNodeRef(_node.tree(), ryml::NONE);
	if (!_index)
		return _node.find_child(key);
	if (!_index->count(key))
		return ryml::ConstNodeRef(_node.tree(), ryml::NONE);
	return _node.tree()->cref(_index->at(key));
}

std::vector<YamlNodeReader> YamlNodeReader::children() const
{
	if (_invalid)
		return std::vector<YamlNodeReader>(0);
	std::vector<YamlNodeReader> children(_node.num_children());
	for (const ryml::ConstNodeRef& child : _node.cchildren())
		children.push_back(YamlNodeReader(child));
	return children;
}

bool YamlNodeReader::isValid() const
{
	return !_invalid;
}

bool YamlNodeReader::isMap() const
{
	return _node.is_map();
}

bool YamlNodeReader::isSeq() const
{
	return _node.is_seq();
}

bool YamlNodeReader::hasVal() const
{
	return _node.has_val();
}

const YamlString YamlNodeReader::getInnerYaml() const
{
	return YamlString(ryml::emitrs_yaml<std::string>(_node));
}

const YamlNodeReader YamlNodeReader::operator[](ryml::csubstr key) const
{
	return YamlNodeReader(getChildNode(key));
}

const YamlNodeReader YamlNodeReader::operator[](size_t pos) const
{
	return YamlNodeReader(_node.child(pos));
}

YamlNodeReader::operator bool() const
{
	return !_invalid;
}

YamlRootNodeReader::YamlRootNodeReader() : YamlNodeReader(), _tree()
{
}

YamlRootNodeReader::YamlRootNodeReader(std::string fullPath) : YamlNodeReader(), _tree(new ryml::Tree())
{
	/*
	size_t size;
	char* data = CrossPlatform::readFileRaw(fullPath, &size);
	ryml::parse_in_arena(ryml::csubstr(data, size), _tree);
	SDL_free(data);
	_node = _tree->crootref();
	_invalid = _node.invalid();
	*/
}

YamlRootNodeReader::YamlRootNodeReader(char* data, size_t size) : YamlNodeReader(), _tree(new ryml::Tree())
{
	/*
	ryml::parse_in_arena(ryml::csubstr(data, size), _tree);
	SDL_free(data);
	_node = _tree->crootref();
	_invalid = _node.invalid();
	*/
}

YamlRootNodeReader::YamlRootNodeReader(const YamlString& yamlString) : YamlNodeReader(), _tree(new ryml::Tree())
{
	ryml::parse_in_arena(ryml::to_csubstr(yamlString.yaml), _tree);
	_node = _tree->crootref();
	_invalid = _node.invalid();
}

YamlRootNodeReader::~YamlRootNodeReader()
{
	delete _tree;
}

YamlNodeWriter::YamlNodeWriter(ryml::NodeRef node) : _node(node)
{
}

YamlNodeWriter YamlNodeWriter::writeKey(ryml::csubstr key)
{
	return YamlNodeWriter(_node.append_child({ryml::KEY, key}));
}

void YamlNodeWriter::setAsMap()
{
	_node |= ryml::MAP;
}

void YamlNodeWriter::setAsSeq()
{
	_node |= ryml::SEQ;
}

void YamlNodeWriter::setFlowStyle()
{
	_node |= ryml::FLOW_SL;
}

void YamlNodeWriter::setBlockStyle()
{
	_node |= ryml::BLOCK;
}

YamlRootNodeWriter::YamlRootNodeWriter() : YamlNodeWriter(_node), _tree(new ryml::Tree)
{
	_node = _tree->rootref();
}

YamlRootNodeWriter::~YamlRootNodeWriter()
{
	delete _tree;
}

std::string YamlRootNodeWriter::emit()
{
	return ryml::emitrs_yaml<std::string>(_tree);
}

}

}
